"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.flattenConfig = exports.applyFlatConfigToState = exports.applyDiffWithCreateQueriesFromPatch = exports.createQueryObject = exports.createQueriesFromPatch = exports.get = exports.formatNamespace = exports.createOrApplyPath = exports.pm = exports.defaultSkipValue = exports.createSearch = void 0;
const shallow_1 = require("zustand/shallow");
const store_js_1 = require("./store.js");
const createSearch = (query) => {
    const queryString = new URLSearchParams(query).toString();
    return queryString === '' ? '' : `?${queryString}`;
};
exports.createSearch = createSearch;
/**
 * Default skip implementation
 */
function defaultSkipValue(value, initialValue) {
    return (value === undefined ||
        value === null ||
        (0, shallow_1.shallow)(value, initialValue) ||
        (Array.isArray(value) && value.length === 0));
}
exports.defaultSkipValue = defaultSkipValue;
function pm(name, serializer, thisSkipValue = defaultSkipValue) {
    return () => ({
        name,
        serializer,
        skipValue: thisSkipValue,
    });
}
exports.pm = pm;
const createOrApplyPath = (obj, path, value = null) => {
    let current = obj || {};
    let thisPath = [...path];
    while (thisPath.length > 1) {
        const [head, ...tail] = thisPath;
        thisPath = tail;
        if (current && current[head] === undefined) {
            current[head] = {};
        }
        current = current === null || current === void 0 ? void 0 : current[head];
    }
    if (current) {
        current[thisPath[0]] = value;
    }
    return obj;
};
exports.createOrApplyPath = createOrApplyPath;
const formatNamespace = (key, ns) => {
    return ns && ns !== store_js_1.DEFAULT_NAMESPACE ? `${ns}.${key}` : key;
};
exports.formatNamespace = formatNamespace;
const get = (object, path) => {
    return path.reduce((next, key) => {
        return next ? next[key] : undefined;
    }, object);
};
exports.get = get;
/**
 * @return a list of patches that deeply match the given config object
 */
const findDeepPatches = (config, basePath) => {
    return Object.keys(config).reduce((next, item) => {
        if (typeof config[item] === 'function') {
            return [...next, { path: [...basePath, item], op: 'replace' }];
        }
        return [
            ...next,
            ...findDeepPatches(config[item], [...basePath, item]),
        ];
    }, []);
};
/**
 * @return an object with the keys that have been processed
 * if a key has been removed / set to undefined, we still return it to
 * be able to create a diff to the current state
 */
const createQueriesFromPatch = (config, ns, patch, state, initialState) => {
    return patch.reduce((next, item) => {
        const { path } = item;
        // namespaces, [ns], values|initialValues, ...rest
        const [, patchNamespace, , ...objectPath] = path;
        // skip patches that don't belong to the given namespace
        if (patchNamespace !== ns) {
            return next;
        }
        const possibleParameter = (0, exports.get)(config, objectPath);
        const isNotCallable = typeof possibleParameter !== 'function';
        if (possibleParameter !== undefined && isNotCallable) {
            // If we have an object as result, we create patches for each parameter inside the subtree
            const patches = findDeepPatches(possibleParameter, path);
            return {
                ...next,
                ...(0, exports.createQueriesFromPatch)(config, ns, patches, state, initialState),
            };
        }
        if (isNotCallable) {
            return next;
        }
        const { name, serializer, skipValue } = possibleParameter();
        const value = (0, exports.get)(state, objectPath);
        const initialValue = (0, exports.get)(initialState, objectPath);
        const nextValue = skipValue(value, initialValue) ? undefined : value;
        return {
            ...next,
            [(0, exports.formatNamespace)(name, ns)]: nextValue === undefined ? nextValue : serializer.serialize(nextValue),
        };
    }, {});
};
exports.createQueriesFromPatch = createQueriesFromPatch;
/**
 * Creates a queryObject that can be serialized.
 */
const createQueryObject = (config, ns, values, initialState) => {
    return Object.keys(config).reduce((next, parameter) => {
        const { path, serializer, skipValue } = config[parameter];
        const possibleValue = (0, exports.get)(values, path);
        const nextValue = skipValue(possibleValue, (0, exports.get)(initialState, path))
            ? undefined
            : possibleValue;
        if (nextValue === undefined) {
            return next;
        }
        return {
            ...next,
            [(0, exports.formatNamespace)(parameter, ns)]: serializer.serialize(nextValue),
        };
    }, {});
};
exports.createQueryObject = createQueryObject;
const applyDiffWithCreateQueriesFromPatch = (config, ns, currentQuery, patch, state, initialState) => {
    const query = (0, exports.createQueriesFromPatch)(config, ns, patch, state, initialState);
    const nextQueries = {
        ...currentQuery,
        ...query,
    };
    return Object.keys(nextQueries)
        .filter((key) => nextQueries[key] !== undefined)
        .reduce((next, key) => {
        return {
            ...next,
            [key]: nextQueries[key],
        };
    }, {});
};
exports.applyDiffWithCreateQueriesFromPatch = applyDiffWithCreateQueriesFromPatch;
/**
 * Applies the given queryValues to the current state, based on the configuration
 * Important: Mutates `state`.
 * @return an object with the keys that have been processed
 */
const applyFlatConfigToState = (config, queryValues, ns, state, initialState, apply = true) => {
    return Object.keys(config).reduce((next, queryParameter) => {
        const { path, serializer, skipValue } = config[queryParameter];
        const nsQueryParameter = (0, exports.formatNamespace)(queryParameter, ns);
        const maybeValue = queryValues[nsQueryParameter];
        const value = maybeValue === undefined
            ? (0, exports.get)(initialState, path)
            : serializer.deserialize(maybeValue);
        if (apply) {
            (0, exports.createOrApplyPath)(state, path, value);
        }
        if (skipValue(value, (0, exports.get)(initialState, path))) {
            return next;
        }
        if (typeof maybeValue === 'undefined') {
            return next;
        }
        return {
            ...next,
            [nsQueryParameter]: maybeValue,
        };
    }, {});
};
exports.applyFlatConfigToState = applyFlatConfigToState;
/**
 * Creates a flat serializable object based on a mapping Object
 * that can be used to generate a query string.
 * @return a key/value object with the object containing the defined parameters as key.
 */
const flattenConfig = (config, path = []) => {
    return Object.keys(config).reduce((next, key) => {
        const v = Array.isArray(config)
            ? config[key]
            : config[key];
        const nextPath = [...path, key];
        if (typeof v === 'function') {
            const { name, ...rest } = v();
            if (next[name] !== undefined) {
                throw new Error(`Config invalid: Multiple definitions found for ${name}.`);
            }
            return {
                ...next,
                [name]: {
                    path: nextPath,
                    ...rest,
                },
            };
        }
        if (typeof v === 'object') {
            return {
                ...next,
                ...(0, exports.flattenConfig)(v, nextPath),
            };
        }
        return next;
    }, {});
};
exports.flattenConfig = flattenConfig;
